/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2013.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

(function(){
dojo.provide("com.ibm.team.vvc.shared.web.ui.util.VvcCache");

dojo.require("dojo.store.Memory");
dojo.require("dojo.store.Cache");
dojo.require("com.ibm.team.vvc.shared.web.ui.util.VvcUtils");
var VvcUtils = com.ibm.team.vvc.shared.web.ui.util.VvcUtils;

/**
 * <p>
 * The VVC cache is a centralized way of getting information about a
 * configuration. Ask it for a configuration by ID, and it will look up all
 * information it knows about already. If it does not know about the
 * configuration, it will ask the server. The VvcCache is actually two
 * separate caches, one responsible for the local data, and one responsible
 * for getting the remote data. Eventually, the entire VVC web UI should be
 * using the VVC cache to get information about configurations.
 * </p>
 * <p>
 * The cache is designed to work with configurations, but can easily be
 * extended to provide information about nearly anything within VVC, if you
 * implement the correct API. The cache has two modes of getting data, a
 * "single" mode, and a "bulk" mode. Single mode is designed for one-off
 * fetches of data from the cache, and is quite uninteresting in that it
 * behaves like any other cache.
 * </p>
 * <p>
 * Bulk mode is where the VvcCache gets interesting. To start a bulk-mode
 * transaction, you simply call "startBulkTransaction" and wait for the
 * deferred to return. <b>Once the deferred returns, the cache is locked and
 * it is ready for a bulk-mode operation, <u>although no other bulk
 * transactions can occur at this time.</u></b> There is no special way to request
 * data when in bulk mode, you proceed as if it were a single request. The
 * cache will automatically aggregate any requests it does not know about.
 * When "endBulkModeTransaction" is called, it will POST a request to the
 * front service asking for all the URI's which were collected during the
 * bulk-mode transaction. It will then resolve the deferreds individually,
 * as if it were a single transaction.
 * </p>
 * <p>
 * To support bulk-mode transactions, there should be a bulk-mode service on
 * the server side which is expecting to receive a list of URI's to resolve.
 * The data coming back from the server should be a well-formed JSON object
 * with key-value pairs. Keys should be the URI's which the cache is
 * requesting, while values should be the corresponding objects. Single-mode
 * transactions should behave the same way.
 * </p>
 * <p>
 * The local cache is a simple HTML5 session store. This allows the cache to
 * survive between page loads; changing pages does not clear the cache. To
 * prevent the cache from going stale, it will automatically kick objects
 * out based on a time limit. Each cache entry has its own time limit,
 * derived either from the cache's default limit, or from some other source.
 * This is achieved through a time stamp. If the current time exceeds this
 * time stamp, the local cache will treat any gets to that object as a cache
 * miss and re-fetch them from the server.
 * </p>
 * <p>
 * To achieve this level of simplicity, the cache is heavily dependent on
 * <code>dojo.Deferred</code>s and callbacks. If you ask the cache for
 * something, it will return a deferred for the request, which will resolve
 * with the item, when the cache knows about the item. Even for a cache hit,
 * a deferred object will be returned. Dojo provides a handy method for
 * dealing with these situations, `<code>dojo.when</code>.
 * <code>dojo.when</code> is designed for situations where you may have a
 * deferred or you may have an object; safely and reliably executing the
 * callback with either the provide object, or the result of the deferred.
 * For more information about <code>dojo.when</code>, please consult the
 * Dojo 1.8 documentation:
 * {@link http://dojotoolkit.org/reference-guide/1.8/dojo/when.html}
 * </p>
 * <p>
 * The cache starts the <code>VvcProxyInterceptor</code> which abstracts the
 * two running modes of the VVC client from the developer.  In cross-domain mode
 * (ex: ConfigurationContextBannerButton), the intercepter will re-write requests
 * bound for VVC core (<code>.../vvc/*</code>) so they use the VVC proxy.  In 
 * core mode (ex: VVC Explorer), the intercepter is not active, and requests run
 * normally.  Developers should always use the cache as if it were running in 
 * core mode.
 * </p>
 * <p>
 * For reference usage, examine
 * <code>com.ibm.team.vvc.server.core.ui.explorer.VvcExplorer</code>.
 * </p>
 */
dojo.declare("com.ibm.team.vvc.shared.web.ui.util.VvcCache", [], {
	
	_localStore: null,
	_idProperty: null,
	_target: null,
	_bulkTarget: null,
	_transaction: null,
	_transactionData: null,
	_locks: null,
	
	constructor: function(args){
		this._localStore = new com.ibm.team.vvc.shared.web.ui.util.VvcLocalCache(args);
		this._idProperty = args.idProperty;
		this._target = args && args.target ? args.target : VvcUtils.getVvcUri() + "/client/single";
		this._bulkTarget = args && args.bulkTarget ? args.bulkTarget : VvcUtils.getVvcUri() + "/client/bulk";
		this._locks = [];
	},
	
	/**
	 * Gets an object from the store based on its id.  If the object
	 * does not exist, a query will goto the server to fetch it.
	 */
	get: function(id, query){
		var cachedResult = this._localStore.get(id);
		if(!cachedResult){
			if(!query){
				query = {
					configurationUri : id
				};
			}
			if(this._isBulkTransaction){
				// Bulk POST mode
				return this._addBulkRequest(query);
			} else {
				// Single GET mode
				return this.query(id, query);
			}
		}
		
		// Remove the "id" property...we know what we're looking for if we've requested it by name already.
		var result = [];
		for(var property in cachedResult){
			if(property !== "id"){
				result[property] = cachedResult[property];
			}
		}
		return result;
	},
	
	/**
	 * Executes a query against the remote data store.  
	 * 
	 * Based on the store#query method.
	 * 
	 * @param The query options.
	 * @return An array or promise.
	 * @see dojo.data.JsonRest
	 */
	query: function(id, query){
		var resultDef;
		if(this._isBulkTransaction){
			resultDef = this._addBulkRequest(query);
		} else {
			resultDef = new dojo.Deferred()
			var args = {
				url: this._target + "?" + dojo.objectToQuery(query),
				headers: {
					"Accept" : "application/json",
					"Content-Type" : "application/json"
				},
				handleAs: "json",
//				preventCache: true
			};
			var xhrDeferred = jazz.client.xhrGet(args);
			xhrDeferred.then(dojo.hitch(this, function(results){
				if(results){
					this.put(results, id);
					resultDef.resolve(results);
				}
			}));
		}
		return resultDef;
	},
	
	/**
	 * Puts an item in the local cache only.  
	 * 
	 * Based on the dojo.store.Memory#put method.
	 * 
	 * @param item The item to put in the cache.
	 * @param properties The properties 
	 */
	put: function(item, id, options){
		this._localStore.put(item, id, options);
	},
	
	/**
	 * Removes an item from the cache.  If no item is specified, 
	 * then the entire cache is cleared.
	 * 
	 * @param id The ID of the item to remove.
	 * 
	 * @return The removed item.
	 */
	clear: function(id){
		return this._localStore.clear(id);
	},
		
	/**
	 * Starts a bulk transaction.  The cache will collect all 
	 * local cache misses and dispatch them at once, resolving 
	 * the individual deferreds one at a time, but only using 
	 * one HTTP request.
	 */
	startBulkTransaction: function(){
		var lock = new dojo.Deferred();
		var locked;
		
		if(this._locks.length > 0){
			locked = this._locks[0];
		}else {
			// Create a dummy lock and release it.
			locked = new dojo.Deferred();
			locked.resolve();
		}
		
		this._locks.unshift(lock);
		dojo.when(locked, dojo.hitch(this, function(){
			this._isBulkTransaction = true;
			this._transactionData = [];
			this._transaction = new dojo.Deferred();
			this._transaction.then(dojo.hitch(this, function(){
				if(this._transactionData.length === 0){
					// Abort if there's nothing to get.
					this._isBulkTransaction = false;
					lock.resolve(true);
					this._locks.splice(this._locks.length - 1, 1);
					return;
				}
				var queries = [];
				dojo.forEach(this._transactionData, dojo.hitch(this, function(item, index, items){
					queries.push(item.requestQuery);
				}));
				var formData = {"queries" : queries};
				var args = {
						url: this._bulkTarget,
						headers: {
							"Accept" : "application/json",
							"Content-Type" : "application/json"
						},
						handleAs: "json",
						postData:  dojo.toJson(formData),
//						preventCache: true
				};
				var post = jazz.client.xhrPost(args);
				post.then(dojo.hitch(this, function(results){
					dojo.forEach(this._transactionData, dojo.hitch(this, function(transaction, index, transactions){
						var uri = transaction.requestQuery.configurationUri;
						if(results[uri]){
							this._isBulkTransaction = false;
							var conf = results[uri];
							this.put(conf, uri);
							transaction.requestDeferred.resolve(conf);
						}
					}));
					lock.resolve(true);
					this._locks.splice(this._locks.length - 1, 1);
				}));
			}));
		}));
		return locked;
	},
	
	/**
	 * Ends a bulk transaction and fires the request to the server.
	 */
	endBulkTransaction: function(transactionId){
		this._transaction.resolve()
	},
	
	_addBulkRequest: function(query){
		var deferred = new dojo.Deferred();
		var args = {
			requestDeferred: deferred,
			requestQuery: query
		};
		this._transactionData.push(args);
		return deferred;
	},
});

dojo.declare("com.ibm.team.vvc.shared.web.ui.util.VvcLocalCache", [], {

	_storageKey: null,
	_data: null,
	_timeout: 60000,	// Timeout in MS.  Default is 1 minute
	_temp: null,
	
	constructor: function(args){			
		this._storageKey = (args && args.storageKey) ? args.storageKey : "VvcSesseionCache-"; 
		if(args){
			this._temp = args.temporary && true; // force boolean
			if(args.timeout){
				this._timeout = args.timeout;
			}
		}
		
		if(!this._temp){
			var value = sessionStorage.getItem(this._storageKey);
			if (value) {
				this._data = dojo.fromJson(value);
			}
		}
				
		if(!this._data){
			this._data = {};
		}

	},
	
	get: function(id){
		var item = this._data[id];
		if(item){
			var now = new Date();
			if(item.expires > now.getTime()){
				return item.data;
			}
		}
		return null;
	},
	
	put: function(data, id, options){
		var now = new Date();
		var expires = now.getTime() + (options && options.timeout ? options.timeout : this._timeout);
		var item = {
			"id": id,
			"data": data,
			"expires": expires
		};
		this._data[id] = item;
		this.flush();
	},
	
	clear: function(id){
		if(id){
			if(this._data[id]){
				this._data[id] = null;
			}
		} else {
			this._data = {};
		}
		this.flush;
	},
	
	flush: function(){
		if(!this._temp){
			var json = dojo.toJson(this._data);
			sessionStorage.setItem(this._storageKey, json);
		}
	}

});
})();
